/*
 * MLE.cpp

 *
 *  Created on: 7 de feb. de 2017
 *      Author: jorge_cmata
 *
 *  Description: This class includes security functions to carry out MLE algorithm
 */


#include <iostream>

#include <cryptopp/cryptlib.h>
#include <cryptopp/sha.h>
#include <cryptopp/osrng.h>
#include <cryptopp/filters.h>
#include <cryptopp/hex.h>
#include <cryptopp/aes.h>
#include <cryptopp/modes.h>
#include <cryptopp/base64.h>
#include <cryptopp/md5.h>
#define CRYPTOPP_ENABLE_NAMESPACE_WEAK 1
#include <cryptopp/md5.h>

#include "aux.h"


using namespace std;
using namespace CryptoPP;

string hexEncode(string input)
{
	string encoded;

	StringSource ss( input, true,
	    new HexEncoder(
	        new StringSink( encoded )
	    ) // HexEncoder
	); // StringSource

	return encoded;
}

string hexDecode(string encoded)
{
	string decoded;

	StringSource ss( encoded, true,
		new HexDecoder(
			new StringSink( decoded )
		) // HexEncoder
	); // StringSource

	return decoded;
}

/**
 * KeyGen: Hash the message received by parameter with SHA256 hash function
 * return the key extracted from the input message itself (32 bytes or 64 hex characters)
 */
string keyGen(string plaintext)
{
	SHA256 sha256;
	byte digest[CryptoPP::SHA256::DIGESTSIZE];
	sha256.CalculateDigest(digest, (const byte*)plaintext.c_str(), plaintext.size());

	CryptoPP::HexEncoder encoder;
	string output;

	encoder.Attach( new CryptoPP::StringSink( output ) );
	encoder.Put( digest, sizeof(digest) );
	encoder.MessageEnd();

	return output;
}

/**
 * TagGen: Hash the ciphertext received by parameter with SHA256 hash function
 * return the hash related with the ciphertext
 */
string tagGen(string key)
{
	SHA256 sha256;
	byte digest[CryptoPP::SHA256::DIGESTSIZE];
	sha256.CalculateDigest(digest, (const byte*)key.c_str(), key.size());

	CryptoPP::HexEncoder encoder;
	string output;

	encoder.Attach( new CryptoPP::StringSink( output ) );
	encoder.Put( digest, sizeof(digest) );
	encoder.MessageEnd();

	return output;
}

/**
 * Generate a random key to use in AES with size 32 bits (Maximum key length)
 */
string generateRandomKey()
{
	AutoSeededRandomPool rnd;

	// Generate a random key
	SecByteBlock key(0x00, AES::MAX_KEYLENGTH);
	rnd.GenerateBlock( key, key.size() );

	CryptoPP::HexEncoder encoder;
	string output;

	encoder.Attach( new CryptoPP::StringSink( output ) );
	encoder.Put( key, key.size() );
	encoder.MessageEnd();
	return output;
}

/**
 * Encrypt: Encrypt the message using AES256. The function needs the Key (string 32 bytes) and the plaintext (string)
 * return the ciphertext obtained
 */
string encrypt(string key, string plaintext)
{
	//Key and IV setup
	string ctrHex = "00000000000000000000000000000000"; //Set a fixed IV (we need the same output from the same input)
	string ctr = hexDecode(ctrHex);
	// Create Cipher Text
	string ciphertext;

	try
	{
		CTR_Mode< AES >::Encryption cbcEncryption;
		cbcEncryption.SetKeyWithIV( (const byte *)key.c_str(), key.size(), (const byte *)ctr.c_str() );

	    StringSource ss( plaintext, true,
	    		new StreamTransformationFilter( cbcEncryption,
	              new StringSink( ciphertext )
	        ) // StreamTransformationFilter
	    ); // StringSource
	}
	catch( const Exception& e )
	{
		cerr << e.what() << endl;
	    exit(1);
	}

	return ciphertext;

}

/**
 * Decrypt: Decrypt the message using AES256. The function needs the Key (string 32 bytes) and the ciphertext (string)
 * return the plaintext
 */
string decrypt(string key, string ciphertext)
{
	//Key and IV setup
	string ctrHex = "00000000000000000000000000000000"; //Set a fixed IV (we need the same output from the same input)
	string ctr = hexDecode(ctrHex);
	// Create Cipher Text
	string plaintext;

	try
	{
	    CTR_Mode< AES >::Decryption cbcDecryption;
	    cbcDecryption.SetKeyWithIV( (const byte *)key.c_str(), key.size(), (const byte *)ctr.c_str() );

	    StringSource ss( ciphertext, true,
	        new StreamTransformationFilter( cbcDecryption,
	            new StringSink( plaintext )
	        ) // StreamTransformationFilter
	    ); // StringSource
	}
	catch( const CryptoPP::Exception& e )
	{
	    cerr << e.what() << endl;
	    exit(1);
	}

	return plaintext;

}

/**
 * Remove Padding: Remove 1+0s at the end of the input string
 * return the new string
 */
string removePadding(string input){
	bool isPadding = false;
	int buffer = 0;
	char lastChar = '0';

	//Loop until the last character will be different of '0'
	while(lastChar == '0'){
		lastChar = input.at(input.size()-1-buffer);
		buffer++;
		isPadding = true;
	}
	//If the padding exits, and now the last char is '1', remove all the padding
	if(isPadding && lastChar=='1'){
		//Remove the padding
		input.erase(input.size()-buffer); //Position where the padding starts
	}
	return input;
}

/**
 * checkRollingHash: Check if a given string accomplishes a certain pattern
 * Hash function: MD5
 * Output: false --> Do not satisfy the condition
 * Output: true --> Satisfy the condition
 */
bool checkRollingHash(string input_text)
{
	Weak1::MD5 md5;
	byte digest[CryptoPP::Weak1::MD5::DIGESTSIZE];
	md5.CalculateDigest(digest, (const byte*)input_text.c_str(), input_text.size());

	for(int i = 1; i<3; i++){
		unsigned char* tempBuffer = &digest[sizeof(digest)-i];
		long binary = (long) *tempBuffer;
		//if the byte is the last one -> Check if char = '0000 0000'
		//If it is the second-to-last one -> Check if byte == 'XXXX 0000'
		if(i==1 && binary!=0){
			return false;
		} else if(i==2 && (binary % 16 != 0)){
			return false;
		}
	}
	return true;
}

/**
 * checkRollingHash: Check if a given string accomplishes a certain pattern
 * Hash function: MD5
 * Output: 0 --> Do not satisfy any condition
 * Output: 1 --> Satisfy soft condition
 * Output: 2 --> Satisfy strong condition
 */
int check2DRollingHash(string input_text)
{
	Weak1::MD5 md5;
	byte digest[CryptoPP::Weak1::MD5::DIGESTSIZE];
	md5.CalculateDigest(digest, (const byte*)input_text.c_str(), input_text.size());

	for(int i = 1; i<3; i++){
		unsigned char* tempBuffer = &digest[sizeof(digest)-i];
		long binary = (long) *tempBuffer;
		//if the byte is the last one -> Check if char = '0000 0000'
		//If it is the second-to-last one -> Check if byte == 'XXXX X000'
		if(i==1 && binary!=0){
			return 0;
		} else if(i==2 && (binary % 8 != 0)){
			return 0;
		} else if(i==2 && (binary % 16 != 0)){
			return 2;
		}
	}
	return 1;
}


/**
 * checkLeapBasedHash: Check if a given string accomplishes a certain pattern
 * Hash function: MD5
 * Output: false --> Do not satisfy the condition
 * Output: true --> Satisfy the condition
 */
bool checkLeapBasedHash(string input_text)
{
	Weak1::MD5 md5;
	byte digest[CryptoPP::Weak1::MD5::DIGESTSIZE];
	md5.CalculateDigest(digest, (const byte*)input_text.c_str(), input_text.size());

	unsigned char* tempBuffer = &digest[sizeof(digest)-1];
	long binary = (long) *tempBuffer;
	if(binary % 2 == 0){
		return true;
	} else if(binary % 16 == 1){
		return true;
	}
	return false;

}



